/// This file implements a lighter alternative version of the Jakob
/// 2019 spectral upsampling method.  Instead of using the entire 3D
/// looking table, we use two 2d slices of the table and interpolate
/// between the evaluated spectral values calculated from those tables.
///
/// The provides similar color matching as full Jakob, at the expense of
/// somewhat lower quality spectrums, and the inability to precalculate
/// the coefficents for even more efficient evaluation later on.
use float4::Float4;

/// How many polynomial coefficients?
const RGB2SPEC_N_COEFFS: usize = 3;

// Include tables generated by the build.rs script
include!(concat!(env!("OUT_DIR"), "/jakob_table_inc.rs"));

#[inline]
pub fn rec709_to_spectrum_p4(lambdas: Float4, rgb: (f32, f32, f32)) -> Float4 {
    small_rgb_to_spectrum_p4(
        REC709_TABLE,
        REC709_TABLE_RES,
        REC709_TABLE_MID_VALUE,
        lambdas,
        rgb,
    )
}

#[inline]
pub fn rec2020_to_spectrum_p4(lambdas: Float4, rgb: (f32, f32, f32)) -> Float4 {
    small_rgb_to_spectrum_p4(
        REC2020_TABLE,
        REC2020_TABLE_RES,
        REC2020_TABLE_MID_VALUE,
        lambdas,
        rgb,
    )
}

#[inline]
pub fn aces_to_spectrum_p4(lambdas: Float4, rgb: (f32, f32, f32)) -> Float4 {
    small_rgb_to_spectrum_p4(
        ACES_TABLE,
        ACES_TABLE_RES,
        ACES_TABLE_MID_VALUE,
        lambdas,
        rgb,
    )
}

//===============================================================
// Core functions, specialized above for specific color spaces.

#[inline(always)]
fn small_rgb_to_spectrum_p4(
    table: &[[(f32, f32, f32); 2]],
    table_res: usize,
    table_mid_value: f32,
    lambdas: Float4,
    rgb: (f32, f32, f32),
) -> Float4 {
    // Determine largest RGB component, and calculate the other two
    // components scaled for lookups.
    let (i, max_val, x, y) = {
        let mut i = 0;
        let mut max_val = rgb.0;
        let mut x = rgb.1;
        let mut y = rgb.2;

        if rgb.1 > max_val {
            i = 1;
            max_val = rgb.1;
            x = rgb.2;
            y = rgb.0;
        }

        if rgb.2 > max_val {
            i = 2;
            max_val = rgb.2;
            x = rgb.0;
            y = rgb.1;
        }

        let scale = 63.0 / max_val;
        x *= scale;
        y *= scale;

        (i, max_val, x, y)
    };

    // Make sure we're not looking up black, to avoid NaN's from divide by zero.
    if max_val == 0.0 {
        return Float4::splat(0.0);
    }

    // Calculate lookup coordinates.
    let xi = (x as usize).min(table_res - 2);
    let yi = (y as usize).min(table_res - 2);
    let offset = (table_res * table_res * i) + (yi * table_res) + xi;
    let dx = 1;
    let dy = table_res;

    // Look up values from table.
    let a0 = table[offset];
    let a1 = table[offset + dx];
    let a2 = table[offset + dy];
    let a3 = table[offset + dy + dx];

    // Convert to SIMD format for faster interpolation.
    let a0 = [
        Float4::new(a0[0].0, a0[0].1, a0[0].2, 0.0),
        Float4::new(a0[1].0, a0[1].1, a0[1].2, 0.0),
    ];
    let a1 = [
        Float4::new(a1[0].0, a1[0].1, a1[0].2, 0.0),
        Float4::new(a1[1].0, a1[1].1, a1[1].2, 0.0),
    ];
    let a2 = [
        Float4::new(a2[0].0, a2[0].1, a2[0].2, 0.0),
        Float4::new(a2[1].0, a2[1].1, a2[1].2, 0.0),
    ];
    let a3 = [
        Float4::new(a3[0].0, a3[0].1, a3[0].2, 0.0),
        Float4::new(a3[1].0, a3[1].1, a3[1].2, 0.0),
    ];

    // Do interpolation.
    let x1: f32 = x - xi as f32;
    let x0: f32 = 1.0 - x1 as f32;
    let y1: f32 = y - yi as f32;
    let y0: f32 = 1.0 - y1 as f32;
    let b0 = [(a0[0] * x0) + (a1[0] * x1), (a0[1] * x0) + (a1[1] * x1)];
    let b1 = [(a2[0] * x0) + (a3[0] * x1), (a2[1] * x0) + (a3[1] * x1)];
    let c = [(b0[0] * y0) + (b1[0] * y1), (b0[1] * y0) + (b1[1] * y1)];

    // Evaluate the spectral function and return the result.
    if max_val <= table_mid_value {
        rgb2spec_eval_4([c[0].get_0(), c[0].get_1(), c[0].get_2()], lambdas)
            * (1.0 / table_mid_value)
            * max_val
    } else if max_val < 1.0 {
        let n = (max_val - table_mid_value) / (1.0 - table_mid_value);
        let s0 = rgb2spec_eval_4([c[0].get_0(), c[0].get_1(), c[0].get_2()], lambdas);
        let s1 = rgb2spec_eval_4([c[1].get_0(), c[1].get_1(), c[1].get_2()], lambdas);
        (s0 * (1.0 - n)) + (s1 * n)
    } else {
        rgb2spec_eval_4([c[1].get_0(), c[1].get_1(), c[1].get_2()], lambdas) * max_val
    }
}

//============================================================
// Coefficient -> eval functions

#[inline(always)]
fn rgb2spec_fma_4(a: Float4, b: Float4, c: Float4) -> Float4 {
    a.fmadd(b, c)
}

fn rgb2spec_eval_4(coeff: [f32; RGB2SPEC_N_COEFFS], lambda: Float4) -> Float4 {
    let co0 = Float4::splat(coeff[0]);
    let co1 = Float4::splat(coeff[1]);
    let co2 = Float4::splat(coeff[2]);

    let x = rgb2spec_fma_4(rgb2spec_fma_4(co0, lambda, co1), lambda, co2);

    let y = Float4::splat(1.0) / (rgb2spec_fma_4(x, x, Float4::splat(1.0))).sqrt();

    rgb2spec_fma_4(Float4::splat(0.5) * x, y, Float4::splat(0.5))
}
